<?php

namespace Drupal\custom_entity_field_updates\Services;

class FuntionService
{
    /**
     * Samples
     *
     * $changesList = [
     * (Update the entity type "custom_entity": install field "new_field")
     * "custom_entity" => [ // Identifier for entity type "custom_entity".
     *   "new_field", // Identifier for field "new_field".
     *   ],
     * ];
     * 
     * (Update the entity type "custom_entity": update field "change_size_field" (change size to 512 keeping data).)
     * "custom_entity" => [ // Identifier for entity type "custom_entity".
     *  "change_size_field" => [ // Identifier for field "change_size_field".
     *    "length" => 512, // Change the property to a this value.
     *  ],
     * ],
     * 
     * (Update the entity type "custom_entity": update field "update_field_keep_data".)
     * 
     * "custom_entity" => [ // Identifier for entity type "custom_entity".
     *   "update_field_keep_data" => [ // Identifier for field "update_field_keep_data".
     *     "keepData" => false, // Do discard all existing data for this field.
     *   ],
     * ],
     * 
     * (Update the entity type "custom_entity": uninstall field "remove_field".)
     * "custom_entity" => [ // Identifier for entity type "custom_entity".
     *   "remove_field", // Identifier for field "remove_field".
     * ],
     */

    /**
     * Summary of changeEntities
     * @param mixed $changesList
     * @return string
     */
    public function changeEntities($changesList)
    {
        // Inspired by https://www.drupal.org/node/2554097
        // Function to update custom entities (from PHP classes to database), v1.1.

        // Declare a structured array referencing fields that need to be changed. This array is
        // structured as $changesList["<entity type identifier>"]["<field identifier>"], where an element
        // at the last level (<identifier of the field>) can be just a label, or its value can be null,
        // an empty array, or an array containing
        //  - the boolean element "keepData" (which defaults to true, when not specified) to tell the
        //    process if it should preserve eventual existing data for that field (this only works with
        //    update operations);
        //  - the scalar element "defaultValue" to assign a value to the field for all records which have
        //    it NULL.

        // Declare empty arrays for return messages.
        $changesDone = [];
        $mistakes = [];

        // Fetch the database connection.
        $database = \Drupal::database();

        // Fetch the entity type manager.
        $entTypeManager = \Drupal::entityTypeManager();
        // Fetch the entity definition update manager.
        $entDefUpdManager = \Drupal::entityDefinitionUpdateManager();

        // Loop for each entity type.
        foreach ($changesList as $entTypeId => $entType) {
            // Get the bundle storage for this entity type.
            $bundleStorage = $entTypeManager->getStorage($entTypeId);
            // Get the bundle definition for this entity type.
            $bundleDef = $entTypeManager->getDefinition($entTypeId);

            // Fetch definitions for all existing fields of this entity type.
            $oldFieldDefs = \Drupal::service(
                "entity.last_installed_schema.repository"
            )->getLastInstalledFieldStorageDefinitions($entTypeId);
            // Fetch definitions for all new fields of this entity type.
            $newFieldDefs = \Drupal::service(
                "entity_field.manager"
            )->getFieldDefinitions($entTypeId, null);

            // Get the identifier of the primary key for this entity type (sometimes the primary key is not
            // labelled "id").
            $primKeyId = $bundleDef->getKey("id");
            // Get the name of the table for this entity type. If there is no data table defined, then use
            // the base table.
            $tableName =
                $bundleStorage->getDataTable() ?:
                $bundleStorage->getBaseTable();

            // Loop for each field to be updated.
            foreach ($entType as $fieldId => $field) {
                // Check if array item for this field is just the label for the identifier.
                if (is_numeric($fieldId)) {
                    // Array item for this field is just the label for the identifier.
                    // Then move the content of variable $field to variable $fieldId, then empty the variable
                    // $field.
                    $fieldId = $field;
                    $field = null;
                }

                // Determine which operation needs to be performed: install ("add"), update ("upd") or
                // uninstall ("rem") the field.
                // Check if field already has an old definition.
                if (isset($oldFieldDefs[$fieldId])) {
                    // An old definition for this field already exists...

                    // Check if field finally has a new definition.
                    if (isset($newFieldDefs[$fieldId])) {
                        // ...a new definition for this field finally exists.
                        // Therefore, this field should only be updated.

                        // Set the update operation.
                        $operation = "upd";
                    } else {
                        // ...there isn't any new definition for this field.
                        // Therefore, this field should be uninstalled.

                        // Set the uninstall operation.
                        $operation = "rem";
                    }
                } else {
                    // There isn't any old definition for this field...

                    // Check if field finally has a new definition.
                    if (isset($newFieldDefs[$fieldId])) {
                        // ...a new definition for this field finally exists.
                        // Therefore, this field should be installed.

                        // Set the install operation.
                        $operation = "add";
                    } else {
                        // ...there isn't any new definition for this field.
                        // There should be a mistake in provided array for changes, because no field can be found
                        // with such label!

                        // Log details about this mistake for the return message.
                        $mistakes[] =
                            "Cannot find field '" .
                            $fieldId .
                            "' of entity type '" .
                            $entTypeId .
                            "'.";

                        // Break the operation for this field.
                        continue;
                    }
                }

                // Determine if field's data should be preserved.
                // Data can be preserved only during updates. If keepData flag is missing, assume it is true.
                $keepData =
                    $operation == "upd" &&
                    (!is_array($field) ||
                        !isset($field["keepData"]) ||
                        $field["keepData"]);

                // Check if field's data should be preserved.
                if ($keepData) {
                    // Field's data should be preserved.

                    // Start a DB transaction.
                    $dbTransaction = $database->startTransaction();

                    // Store existing values into a temporary structured array (primary key => field value, as
                    // per the "fetchAllKeyed" method).
                    $storedData = $database
                        ->select($tableName)
                        ->fields($tableName, [$primKeyId, $fieldId])
                        ->execute()
                        ->fetchAllKeyed();
                }

                // Check if operation is update or uninstall.
                if ($operation == "upd" || $operation == "rem") {
                    // Operation is update or uninstall.
                    // Remove existing field definition.

                    // Check if field's table can be truncated.
                    if (
                        !is_array($field) ||
                        !isset($field["truncateTable"]) ||
                        $field["truncateTable"]
                    ) {
                        // Field's table can be truncated.

                        // Clear out all existing values.
                        $database
                            ->update($tableName)
                            ->fields([$fieldId => null])
                            ->execute();
                    }

                    // Fetch the existing field storage definition.
                    $fieldStorageDef = $entDefUpdManager->getFieldStorageDefinition(
                        $fieldId,
                        $entTypeId
                    );
                    // Uninstall the field.
                    $entDefUpdManager->uninstallFieldStorageDefinition(
                        $fieldStorageDef
                    );
                }

                // Check if operation is install or update.
                // if($operation == "add") {
                if ($operation == "add" || $operation == "upd") {
                    // Operation is install or update.
                    // Add new field definition.

                    // Install the new field definition.
                    $entDefUpdManager->installFieldStorageDefinition(
                        $fieldId,
                        $entTypeId,
                        $entTypeId,
                        $newFieldDefs[$fieldId]
                    );
                }

                // Check if field's data should have been preserved.
                if ($keepData) {
                    // Field's data should have been preserved.

                    // Restore previous values.
                    // Loop for each existing value (i.e. each entity record).
                    foreach ($storedData as $primKey => $fieldValue) {
                        // Write the existing values into the table.
                        $database
                            ->update($tableName)
                            ->fields([$fieldId => $fieldValue])
                            ->condition($primKeyId, $primKey)
                            ->execute();
                    }
                }

                // Check if operation is install or update.
                if ($operation == "add" || $operation == "upd") {
                    // Operation is install or update.

                    // Try to get the default value literal from the new definition of field.
                    $defaultValueLiteral = $newFieldDefs[
                        $fieldId
                    ]->getDefaultValueLiteral();

                    // Check if default value was declared in the list of changes. If so, this will have
                    // precedence on same setting from new definition of field.
                    if (is_array($field) && isset($field["defaultValue"])) {
                        $defaultValue = $field["defaultValue"];
                    }
                    // Otherwise, check if the new definition of field contains a default value.
                    elseif (
                        is_array($defaultValueLiteral) &&
                        count($defaultValueLiteral) > 0
                    ) {
                        // Take the first element from array "default value literal", then its "value" item.
                        $defaultValue = $defaultValueLiteral[0]["value"];
                    } else {
                        // Unset the default value.
                        $defaultValue = null;
                    }

                    // Check if a default value should be set.
                    if (!is_null($defaultValue)) {
                        // A default value should be set.
                        // Set the default value to all records which currently have null values.

                        // Write the default value into the table (only to records which have it null).
                        $defaultValuesApplied = $database
                            ->update($tableName)
                            ->fields([$fieldId => $defaultValue])
                            ->isNull($fieldId)
                            ->execute();
                    }
                }

                // Check if field's data should have been preserved.
                if ($keepData) {
                    // Field's data should have been preserved.

                    // Commit the DB transaction.
                    unset($dbTransaction);
                }

                // Log details about current operation for the return message.
                $changesDone[] =
                    ($operation == "add"
                        ? "Installed"
                        : ($operation == "upd"
                            ? "Updated"
                            : "Uninstalled")) .
                    " field '" .
                    $fieldId .
                    "' of entity type '" .
                    $entTypeId .
                    "'" .
                    ($keepData
                        ? " (" .
                            count($storedData) .
                            " record" .
                            (count($storedData) != 1 ? "s" : "") .
                            " found" .
                            (count($storedData) > 0 ? " and preserved" : "") .
                            ")"
                        : "") .
                    (isset($defaultValue) && !is_null($defaultValue)
                        ? " (default value \"" .
                            $defaultValue .
                            "\" has been set to " .
                            $defaultValuesApplied .
                            " record" .
                            ($defaultValuesApplied != 1 ? "s" : "") .
                            ")"
                        : "") .
                    ".";
            }
        }

        // Check if at least one field was changed, or if there was at least one mistake.
        if (count($changesDone) > 0 || count($mistakes) > 0) {
            // At least one field was updated, or if there was at least one mistake.

            // Return a message.
            return "Update function performed " .
                count($changesDone) .
                " operation" .
                (count($changesDone) != 1 ? "s" : "") .
                "." .
                (count($changesDone) > 0 ? "\n\t    - " : "") .
                implode("\n\t    - ", $changesDone) .
                (count($mistakes) > 0
                    ? "\n [warn]   " .
                        count($mistakes) .
                        " mistake" .
                        (count($mistakes) > 1 ? "s" : "") .
                        " found in the list of changes provided at the beginning of this update function!" .
                        "\n\t    - " .
                        implode("\n\t    - ", $mistakes)
                    : "");
        } else {
            // No fields have been updated and there were no mistakes, i.e. the list of changes was empty.

            // Return a message.
            return " [warn]   No operations performed! Please check the list of changes provided at the" .
                " beginning of this update function.";
        }
    }

}
